------------------------------------------------------------------------
-- Subtracting from a vector
------------------------------------------------------------------------

module LazyVec.VecSubtract where

open import Relation.Binary.PropositionalEquality
open import Data.Function
open import Data.Nat
open import Data.Nat.Properties
open import Data.Vec

infixl 6 _⊖_

_⊖_ : forall {n} -> Vec ℕ n -> ℕ -> Vec ℕ n
[]       ⊖ n = []
(x ∷ xs) ⊖ n = (x ∸ n) ∷ xs ⊖ (n ∸ x)

abstract
  mutual
    xs⊖0≡xs : forall {n} -> (xs : Vec ℕ n) -> xs ⊖ 0 ≡ xs
    xs⊖0≡xs []       = ≡-refl
    xs⊖0≡xs (x ∷ xs) = ≡-cong (_∷_ x) (xs⊖0∸x≡xs x xs)

    xs⊖0∸x≡xs : forall {n} x -> (xs : Vec ℕ n) -> xs ⊖ (0 ∸ x) ≡ xs
    xs⊖0∸x≡xs x xs = ≡-subst (\n -> xs ⊖ n ≡ xs)
                             (≡-sym $ 0∸n≡0 x)
                             (xs⊖0≡xs xs)

  xs⊖1⊖x≡xs⊖1+x :  forall {n} x -> (xs : Vec ℕ n)
                -> xs ⊖ 1 ⊖ x ≡ xs ⊖ (1 + x)
  xs⊖1⊖x≡xs⊖1+x x []           = ≡-refl
  xs⊖1⊖x≡xs⊖1+x x (zero  ∷ ys) =
    ≡-cong₂ _∷_ (0∸n≡0 x) (xs⊖1⊖x≡xs⊖1+x x ys)
  xs⊖1⊖x≡xs⊖1+x x (suc y ∷ ys) =
    ≡-cong (_∷_ (y ∸ x))
           (≡-cong (\ys -> ys ⊖ (x ∸ y)) (xs⊖0∸x≡xs y ys))
