-- # import nat
-- # import digit
-- # import prelude
-- # import tree23
-- # import thunk
-- # import debitPassing

------------------------------------------------------------------------
-- Running times for many functions on finger trees. They depend on
-- each other in various ways, which may not match their order of
-- definition, so they have been collected here.

viewL'Steps : ℕ
viewL'Steps = 1 + headDigitSteps + tailDigitSteps

viewR'Steps : ℕ
viewR'Steps = 1 + lastDigitSteps 3 + initDigitSteps 3

deepLHelperSteps : ℕ
deepLHelperSteps = 1

deepLSteps : ℕ
deepLSteps = 1 + deepLHelperSteps + viewL'Steps

deepRSteps : ℕ
deepRSteps = 1 + deepLHelperSteps + viewR'Steps

-- The value of safeSteps is derived from prop_deepLSteps,
-- prop_deepRSteps and prop_cons'Steps.

safeSteps : ℕ
safeSteps = deepLSteps `max` deepRSteps `max` 2

prop_deepLSteps : deepLSteps ≤ safeSteps
prop_deepLSteps = plusIncreasingRight

prop_deepRSteps : deepRSteps ≤ safeSteps
prop_deepRSteps = plusIncreasingRight

viewLSteps : ℕ
viewLSteps = viewL'Steps + deepLSteps + safeSteps

viewRSteps : ℕ
viewRSteps = viewR'Steps + deepRSteps + safeSteps

-- The value of cons'Steps is derived from prop_cons'Steps and
-- prop_cons'Steps'.

cons'Steps : ℕ
cons'Steps = 1 `max` safeSteps

cons'HelperSteps : ℕ
cons'HelperSteps = (cons'Steps + 1) `max` safeSteps

-- This property arises from the definition of (◃′).

prop_cons'Steps : cons'Steps + safeSteps ≥ 1 + cons'HelperSteps
prop_cons'Steps = plusIncreasingRight

-- This property arises from the definition of consHelper (the case
-- where l has length 4 and r is Safe).

prop_cons'Steps' : cons'Steps ≥ safeSteps
prop_cons'Steps' = plusIncreasingRight

consSteps : ℕ
consSteps = cons'Steps + safeSteps

-- Type synonym used to make the debit passing analysis easier to
-- follow.

FTThunk : ℕ -> ℕ -> Set -> Set
FTThunk = DPThunk safeSteps

------------------------------------------------------------------------
-- Specialised finger tree digits

-- Digits with two or three elements are safe.

data IsSafe : Set where
  Safe   : IsSafe
  Unsafe : IsSafe

isSafe : (len : ℕ) -> IsSafe
isSafe 0                               = Unsafe
isSafe 1                               = Unsafe
isSafe 2                               = Safe
isSafe 3                               = Safe
isSafe 4                               = Unsafe
isSafe (suc (suc (suc (suc (suc _))))) = Unsafe

toSteps : IsSafe -> ℕ
toSteps Safe   = safeSteps
toSteps Unsafe = 0

prop_toSteps (s : IsSafe) : toSteps s ≤ safeSteps
prop_toSteps {Safe}   = plusIncreasingRight
prop_toSteps {Unsafe} = plusIncreasingRight

prop_toSteps₂ (s₁, s₂ : IsSafe) : toSteps s₁ + toSteps s₂ ≤ 2 * safeSteps
prop_toSteps₂ = plusMon prop_toSteps prop_toSteps

toSteps' : ℕ -> ℕ
toSteps' = toSteps ∘ isSafe

data DigitFT' (a : Set) : IsSafe -> Set where
  digitFT (len : ℕ)
    : Digit 1 (len + 1) 4 a -> DigitFT' a (isSafe (len + 1))

DigitFT : IsSafe -> Set -> Set
DigitFT s a = DigitFT' a s

------------------------------------------------------------------------
-- 2-3 finger trees

data FingerTree' (a : Set) : (above : ℕ) -> Set where
  Empty (above : ℕ) : FingerTree' a above
  Single (above : ℕ)
    : Tree23 a above -> FingerTree' a above
  Deep (safeL, safeR : IsSafe) (above : ℕ)
    :  DigitFT safeL (Tree23 a above)
    -> DigitFT safeR (Tree23 a above)
    -> Thunk (toSteps safeL + toSteps safeR)
             (FingerTree' a (above + 1))
    -> FingerTree' a above

FingerTree : (a : Set) -> Set
FingerTree a = FingerTree' a 0

depthFT (above : ℕ) (a : Set) : FingerTree' a above -> ℕ
depthFT Empty        = 0
depthFT (Single _)   = 0
depthFT (Deep _ _ m) = 1 + depthFT (force m)

exampleFT : Thunk 0 (FingerTree ℕ)
exampleFT = return $
  Deep d12 d1214 $ return' $ Deep d36 d711 $ return' Empty
  where
  d12   = digitFT $ digit¹₂ (Zero 1) (Zero 2)
  d36   = digitFT $ digit¹₂ (Succ2 (Zero 3) (Zero 4))
                            (Succ2 (Zero 5) (Zero 6))
  d711  = digitFT $ digit¹₂ (Succ3 (Zero 7) (Zero 8) (Zero 9))
                            (Succ2 (Zero 10) (Zero 11))
  d1214 = digitFT $ digit¹₃ (Zero 12) (Zero 13) (Zero 14)
