-- # import thunk
-- # import debitPassing
-- # import nat
-- # import order
-- # import thunkedPair
-- # import thunkedMaybe

------------------------------------------------------------------------
-- Cost analysis

addSteps        = 1
lviewSteps      = 1

-- Exact cost of a debit in this file, determined from the properties
-- below.

debitSteps = addSteps `max` lviewSteps

prop_addSteps : addSteps ≤ debitSteps
prop_addSteps = plusIncreasingRight

prop_lviewSteps : lviewSteps ≤ debitSteps
prop_lviewSteps = plusIncreasingRight

QThunk  = DPThunk debitSteps
QThunk' = QThunk 0

------------------------------------------------------------------------
-- Queue data type

data Queue (a : Set) : Set where
  Empty   : Queue a
  Single  : a -> Queue a
  TwoZero : a -> a -> QThunk' 2 (Queue (a × a))      -> Queue a
  OneZero : a ->      QThunk' 1 (Queue (a × a))      -> Queue a
  TwoOne  : a -> a -> QThunk' 1 (Queue (a × a)) -> a -> Queue a
  OneOne  : a ->                 Queue (a × a)  -> a -> Queue a

------------------------------------------------------------------------
-- Snoc

-- Discharges one debit, unshared cost: addSteps.

add (a : Set) : Queue a -> a -> QThunk addSteps 1 (Queue a)
add Empty              a = return' $ Single a
add (Single a)         b = return' $ TwoZero a b (return' Empty)
add (TwoZero a b csC)  d = thunkM.do
  cs <- pay debitSteps csC
  return' $ TwoOne a b cs d
add (OneZero a bsC)    c = thunkM.do
  bs <- bsC
  return' $ OneOne a bs c
add (TwoOne a b csC d) e = step $
  csC `seqBind` \cs ->
  return $ TwoZero a b (add cs $ pair d e)
add (OneOne a bs c)    d = step $
  return (OneZero a) `ap` pay debitSteps (add bs $ pair c d)

------------------------------------------------------------------------
-- Left view data type

data LView (a : Set) : Set where
  Nil  : LView a
  Cons : a -> QThunk lviewSteps 1 (Queue a) -> LView a

------------------------------------------------------------------------
-- Left view

-- Discharges zero debits, unshared cost: lviewSteps.

lview (a : Set) : Queue a -> Thunk lviewSteps (LView a)
lview Empty            = return' Nil
lview (Single a)       = return' $ Cons a (return' Empty)
lview (TwoZero a b cs) = return' $
  Cons a (return' (OneZero b) `ap` pay debitSteps cs)
lview {α} (OneZero a bsC) = step $ return $
  Cons a (expand =<< lview =<< bsC)
  where
  expand :  LView (α × α) -> Thunk 0 (Queue α)
  expand Nil                     = return $ Empty
  expand (Cons (pair ba bb) bcs) = return $ TwoZero ba bb bcs
lview (TwoOne a b csC d) = return' $ Cons a $ thunkM.do
  cs <- csC
  return' $ OneOne b cs d
lview {α} (OneOne a bs c) = step $ return $
    Cons a (expand =<< lview bs)
  where
  expand :  LView (α × α) -> QThunk 0 1 (Queue α)
  expand Nil                      = return' $ Single c
  expand (Cons (pair ba bb) bcsC) = thunkM.do
    bcs <- pay debitSteps bcsC
    return $ TwoOne ba bb bcs c
