------------------------------------------------------------------------
-- Proofs showing that erasure commutes with almost anything
------------------------------------------------------------------------

module Erasure.Lemmas where

open import Data.Nat
open import Data.List using ([]; _∷_)
open import Data.Function
open import TypeSystem
open import Erasure
open import OperationalSemantics.Binding
open import Value
open import Substitution
open VarSubst
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning

------------------------------------------------------------------------
-- Main lemmas

-- Some definitions are commented out below. They currently fail due
-- to an Agda bug, and have been replaced by more verbose definitions.
-- No cases are missing.

abstract

  ⌜wk⇒⌝-lemma :  forall {Γ Δ} σ (ρ : Γ ⇒ Δ)
              -> wk⇒ ⌜ ρ ⌝⇒ ≡ ⌜ wk⇒ {σ = σ} ρ ⌝⇒
  ⌜wk⇒⌝-lemma σ ∅       = byDef
  ⌜wk⇒⌝-lemma σ (ρ ▷ x) = begin
    wk⇒ ⌜ ρ ▷ x ⌝⇒
                                            ≡⟨ byDef ⟩
    wk⇒ ⌜ ρ ⌝⇒ ▷ vs ⌜ x ⌝∋
                                            ≡⟨ ⌜wk⇒⌝-lemma σ ρ
                                                 ⟨ ≡-cong₂ _▷_ ⟩
                                               byDef ⟩
    ⌜ wk⇒ {σ = σ} ρ ⌝⇒ ▷ ⌜ vs {σ = σ} x ⌝∋
                                            ≡⟨ byDef ⟩
    ⌜ wk⇒ {σ = σ} (ρ ▷ x) ⌝⇒
                                            ∎

  ⌜↑⌝-lemma :  forall {Γ Δ} σ (ρ : Γ ⇒ Δ)
            -> ⌜ ρ ⌝⇒ ↑ ≡ ⌜ _↑ {σ = σ} ρ ⌝⇒
  ⌜↑⌝-lemma σ ρ = ≡-cong₂ _▷_ (⌜wk⇒⌝-lemma σ ρ) byDef

  ⌜id⌝-lemma : forall Γ -> idˢ ≡ ⌜ idˢ {Γ = Γ} ⌝⇒
  ⌜id⌝-lemma ε       = byDef
  ⌜id⌝-lemma (Γ ▻ σ) =
                                   begin
    idˢ
                                   ≡⟨ byDef ⟩
    idˢ ↑
                                   ≡⟨ ≡-cong _↑ (⌜id⌝-lemma Γ) ⟩
    ⌜ idˢ {Γ = Γ} ⌝⇒ ↑
                                   ≡⟨ ⌜↑⌝-lemma σ (idˢ {Γ = Γ}) ⟩
    ⌜ _↑ {σ = σ} (idˢ {Γ = Γ}) ⌝⇒
                                   ≡⟨ byDef ⟩
    ⌜ idˢ {Γ = Γ ▻ σ} ⌝⇒
                                   ∎

  ⌜wk⌝-lemma : forall Γ σ -> wk ≡ ⌜ wk {Γ = Γ} {σ = σ} ⌝⇒
  ⌜wk⌝-lemma Γ σ =
                                    begin
    wk
                                    ≡⟨ byDef ⟩
    wk⇒ idˢ
                                    ≡⟨ ≡-cong wk⇒ (⌜id⌝-lemma Γ) ⟩
    wk⇒ ⌜ idˢ {Γ = Γ} ⌝⇒
                                    ≡⟨ ⌜wk⇒⌝-lemma σ (idˢ {Γ = Γ}) ⟩
    ⌜ wk⇒ {σ = σ} (idˢ {Γ = Γ}) ⌝⇒
                                    ≡⟨ byDef ⟩
    ⌜ wk {Γ = Γ} {σ = σ} ⌝⇒
                                    ∎

  ⌜sub⌝-lemma : forall {Γ σ} (x : Γ ∋ σ) -> sub ⌜ x ⌝∋ ≡ ⌜ sub x ⌝⇒
  ⌜sub⌝-lemma {Γ = Γ} x = ≡-cong₂ _▷_ (⌜id⌝-lemma Γ) byDef

  ⌜/∋⌝-lemma :  forall {Γ Δ σ} (x : Γ ∋ σ) (ρ : Γ ⇒ Δ)
             -> ⌜ x ⌝∋ /∋ ⌜ ρ ⌝⇒ ≡ ⌜ x /∋ ρ ⌝∋
  ⌜/∋⌝-lemma vz                               (ρ ▷ t) = byDef
  ⌜/∋⌝-lemma (vs v)                           (ρ ▷ t) = ⌜/∋⌝-lemma v ρ
  ⌜/∋⌝-lemma (gwait∋ {σ = τ} {n₁ = n₁} C m x) ρ
    with ⌜ C [ Thunk (m + n₁) τ ] ⌝⋆
       | erase-lemma {m = n₁} {n = m + n₁} {τ = τ} C
  ... | ._ | ≡-refl = ⌜/∋⌝-lemma x ρ

  ⌜∘⌝-lemma :  forall {Γ Δ X} (ρ₁ : Γ ⇒ Δ) (ρ₂ : Δ ⇒ X)
            -> ⌜ ρ₁ ⌝⇒ ∘ˢ ⌜ ρ₂ ⌝⇒ ≡ ⌜ ρ₁ ∘ˢ ρ₂ ⌝⇒
  ⌜∘⌝-lemma ∅        ρ₂ = byDef
  ⌜∘⌝-lemma (ρ₁ ▷ x) ρ₂ =
                                                begin
    ⌜ ρ₁ ▷ x ⌝⇒ ∘ˢ ⌜ ρ₂ ⌝⇒
                                                ≡⟨ byDef ⟩
    (⌜ ρ₁ ⌝⇒ ▷ ⌜ x ⌝∋) ∘ˢ ⌜ ρ₂ ⌝⇒
                                                ≡⟨ byDef ⟩
    (⌜ ρ₁ ⌝⇒ ∘ˢ ⌜ ρ₂ ⌝⇒) ▷ (⌜ x ⌝∋ /∋ ⌜ ρ₂ ⌝⇒)
                                                ≡⟨ ≡-cong₂ _▷_ (⌜∘⌝-lemma ρ₁ ρ₂)
                                                               (⌜/∋⌝-lemma x ρ₂) ⟩
    ⌜ ρ₁ ∘ˢ ρ₂ ⌝⇒ ▷ ⌜ x /∋ ρ₂ ⌝∋
                                                ≡⟨ byDef ⟩
    ⌜ ρ₁ ∘ˢ ρ₂ ▷ x /∋ ρ₂ ⌝⇒
                                                ∎

  ⌜/⊢⌝-lemma :  forall {Γ Δ σ} (e : Γ ⊢ σ) (ρ : Γ ⇒ Δ)
             -> ⌜ e ⌝ /⊢ ⌜ ρ ⌝⇒ ≡ ⌜ e /⊢ ρ ⌝
  ⌜/⊢⌝-lemma (var x)        ρ = ≡-cong var (⌜/∋⌝-lemma x ρ)
  ⌜/⊢⌝-lemma (λ_ {σ = σ} e) ρ =
                                    begin
    (λ ⌜ e ⌝) /⊢ ⌜ ρ ⌝⇒
                                    ≡⟨ byDef ⟩
    λ (⌜ e ⌝ /⊢ ⌜ ρ ⌝⇒ ↑)
                                    ≡⟨ ≡-cong λ_ (≡-cong₂ _/⊢_ (byDef {x = ⌜ e ⌝})
                                                               (⌜↑⌝-lemma σ ρ)) ⟩
    λ (⌜ e ⌝ /⊢ ⌜ _↑ {σ = σ} ρ ⌝⇒)
                                    ≡⟨ ≡-cong λ_ (⌜/⊢⌝-lemma e (_↑ {σ = σ} ρ)) ⟩
    λ ⌜ e /⊢ _↑ {σ = σ} ρ ⌝
                                    ≡⟨ byDef ⟩
    ⌜ λ_ {σ = σ} (e /⊢ ρ ↑) ⌝
                                    ≡⟨ byDef ⟩
    ⌜ λ e /⊢ ρ ⌝
                                    ∎
  ⌜/⊢⌝-lemma (λ²_ {σ₁ = σ₁} {σ₂ = σ₂} e) ρ =
                                                    begin
    λ² (⌜ e ⌝ /⊢ ⌜ ρ ⌝⇒ ↑ ↑)
                                                    ≡⟨ ≡-cong λ²_ (≡-cong₂ _/⊢_ (byDef {x = ⌜ e ⌝})
                                                                                (≡-cong _↑ (⌜↑⌝-lemma σ₁ ρ))) ⟩
    λ² (⌜ e ⌝ /⊢ ⌜ _↑ {σ = σ₁} ρ ⌝⇒ ↑)
                                                    ≡⟨ ≡-cong λ²_ (≡-cong₂ _/⊢_ (byDef {x = ⌜ e ⌝})
                                                                                (⌜↑⌝-lemma σ₂ (_↑ {σ = σ₁} ρ))) ⟩
    λ² (⌜ e ⌝ /⊢ ⌜ _↑ {σ = σ₂} (_↑ {σ = σ₁} ρ) ⌝⇒)
                                                    ≡⟨ ≡-cong λ²_ (⌜/⊢⌝-lemma e (ρ ↑ ↑)) ⟩
    λ² ⌜ e /⊢ ρ ↑ ↑ ⌝
                                                    ∎
  ⌜/⊢⌝-lemma (fix {n = n} {σ = σ} e) ρ =
                                       begin
    fix (⌜ e ⌝ /⊢ ⌜ ρ ⌝⇒ ↑)
                                       ≡⟨ ≡-cong fix (≡-cong₂ _/⊢_ (byDef {x = ⌜ e ⌝})
                                                                   (⌜↑⌝-lemma σ' ρ)) ⟩
    fix (⌜ e ⌝ /⊢ ⌜ _↑ {σ = σ'} ρ ⌝⇒)
                                       ≡⟨ ≡-cong fix (⌜/⊢⌝-lemma e (_↑ {σ = σ'} ρ)) ⟩
    fix ⌜ e /⊢ _↑ {σ = σ'} ρ ⌝
                                       ∎
    where σ' = Thunk (suc n) σ
  ⌜/⊢⌝-lemma (e₁ · e₂) ρ = ≡-cong₂ _·_ (⌜/⊢⌝-lemma e₁ ρ) (⌜/⊢⌝-lemma e₂ ρ)
  ⌜/⊢⌝-lemma z         ρ = byDef
  ⌜/⊢⌝-lemma s         ρ = byDef
  ⌜/⊢⌝-lemma (natrec {n = n} {σ = σ} e₁ e₂) ρ =
                                                    begin
    f (⌜ e₂ ⌝ /⊢ ⌜ ρ ⌝⇒ ↑ ↑)
                                                    ≡⟨ ≡-cong f (≡-cong₂ _/⊢_ (byDef {x = ⌜ e₂ ⌝})
                                                                              (≡-cong _↑ (⌜↑⌝-lemma σ₁ ρ))) ⟩
    f (⌜ e₂ ⌝ /⊢ ⌜ _↑ {σ = σ₁} ρ ⌝⇒ ↑)
                                                    ≡⟨ ≡-cong f (≡-cong₂ _/⊢_ (byDef {x = ⌜ e₂ ⌝})
                                                                              (⌜↑⌝-lemma σ₂ (_↑ {σ = σ₁} ρ))) ⟩
    f (⌜ e₂ ⌝ /⊢ ⌜ _↑ {σ = σ₂} (_↑ {σ = σ₁} ρ) ⌝⇒)
                                                    ≡⟨ ≡-cong₂ natrec (⌜/⊢⌝-lemma e₁ ρ)
                                                                      (⌜/⊢⌝-lemma e₂ (ρ ↑ ↑)) ⟩
    natrec ⌜ e₁ /⊢ ρ ⌝ ⌜ e₂ /⊢ ρ ↑ ↑ ⌝
                                                    ∎
    where
    σ₁ = Nat
    σ₂ = Thunk (suc n) σ
    f  = natrec (⌜ e₁ ⌝ /⊢ ⌜ ρ ⌝⇒)
  ⌜/⊢⌝-lemma (e₁ , e₂) ρ = ≡-cong₂ _,_ (⌜/⊢⌝-lemma e₁ ρ) (⌜/⊢⌝-lemma e₂ ρ)
  ⌜/⊢⌝-lemma (uncurry {σ₁ = σ₁} {σ₂ = σ₂} e) ρ =
                                                         begin
    uncurry (⌜ e ⌝ /⊢ ⌜ ρ ⌝⇒ ↑ ↑)
                                                         ≡⟨ ≡-cong uncurry (≡-cong₂ _/⊢_ (byDef {x = ⌜ e ⌝})
                                                                                         (≡-cong _↑ (⌜↑⌝-lemma σ₁ ρ))) ⟩
    uncurry (⌜ e ⌝ /⊢ ⌜ _↑ {σ = σ₁} ρ ⌝⇒ ↑)
                                                         ≡⟨ ≡-cong uncurry (≡-cong₂ _/⊢_ (byDef {x = ⌜ e ⌝})
                                                                                         (⌜↑⌝-lemma σ₂ (_↑ {σ = σ₁} ρ))) ⟩
    uncurry (⌜ e ⌝ /⊢ ⌜ _↑ {σ = σ₂} (_↑ {σ = σ₁} ρ) ⌝⇒)
                                                         ≡⟨ ≡-cong uncurry (⌜/⊢⌝-lemma e (ρ ↑ ↑)) ⟩
    uncurry ⌜ e /⊢ ρ ↑ ↑ ⌝
                                                         ∎
  ⌜/⊢⌝-lemma (return e)  ρ = ⌜/⊢⌝-lemma e ρ
  ⌜/⊢⌝-lemma (e₁ >>= e₂) ρ = ≡-cong₂ _·_ (⌜/⊢⌝-lemma e₂ ρ)
                                         (⌜/⊢⌝-lemma e₁ ρ)
  ⌜/⊢⌝-lemma (√ e)       ρ = ⌜/⊢⌝-lemma e ρ
  ⌜/⊢⌝-lemma (pay m e)   ρ = ⌜/⊢⌝-lemma e ρ
  ⌜/⊢⌝-lemma (gpay {σ = τ} {n₁ = n₁} C m e) ρ
    with ⌜ C [ Thunk (n₁ ∸ m) τ ] ⌝⋆
       | erase-lemma {m = n₁} {n = n₁ ∸ m} {τ = τ} C
  ... | ._ | ≡-refl = ⌜/⊢⌝-lemma e ρ
  ⌜/⊢⌝-lemma (gwait {σ = τ} {n₁ = n₁} C m e) ρ
    with ⌜ C [ Thunk (m + n₁) τ ] ⌝⋆
       | erase-lemma {m = n₁} {n = m + n₁} {τ = τ} C
  ... | ._ | ≡-refl = ⌜/⊢⌝-lemma e ρ

  ⌜/↦⌝-lemma :  forall {Γ Δ σ} (e : Γ ↦ σ) (ρ : Γ ⇒ Δ)
             -> ⌜ e ⌝↦ /↦ ⌜ ρ ⌝⇒ ≡ ⌜ e /↦ ρ ⌝↦
  ⌜/↦⌝-lemma (↦ⁿˢ {τ = τ} ns e) ρ
    with ⌜ τ ⌝⋆ | Thunks-lemma ns {τ = τ}
  ... | ._ | ≡-refl = ≡-cong ↦_ (⌜/⊢⌝-lemma e ρ)

  ⌜/⊢⇓⌝-lemma :  forall {Γ Δ σ} (v : Γ ⊢⇓ σ) (ρ : Γ ⇒ Δ)
              -> ⌜ v ⌝⇓ /⊢⇓ ⌜ ρ ⌝⇒ ≡ ⌜ v /⊢⇓ ρ ⌝⇓
  ⌜/⊢⇓⌝-lemma (λ⇓_ {σ = σ} e) ρ =
                                     begin
    λ⇓ (⌜ e ⌝ /⊢ ⌜ ρ ⌝⇒ ↑)
                                     ≡⟨ ≡-cong λ⇓_ (≡-cong₂ _/⊢_ (byDef {x = ⌜ e ⌝})
                                                                 (⌜↑⌝-lemma σ ρ)) ⟩
    λ⇓ (⌜ e ⌝ /⊢ ⌜ _↑ {σ = σ} ρ ⌝⇒)
                                     ≡⟨ ≡-cong λ⇓_ (⌜/⊢⌝-lemma e (_↑ {σ = σ} ρ)) ⟩
    λ⇓ ⌜ e /⊢ _↑ {σ = σ} ρ ⌝
                                     ∎
  ⌜/⊢⇓⌝-lemma (λ²⇓_ {σ₁ = σ₁} {σ₂ = σ₂} e) ρ =
                                                     begin
    λ²⇓ (⌜ e ⌝ /⊢ ⌜ ρ ⌝⇒ ↑ ↑)
                                                     ≡⟨ ≡-cong λ²⇓_ (≡-cong₂ _/⊢_ (byDef {x = ⌜ e ⌝})
                                                                                  (≡-cong _↑ (⌜↑⌝-lemma σ₁ ρ))) ⟩
    λ²⇓ (⌜ e ⌝ /⊢ ⌜ _↑ {σ = σ₁} ρ ⌝⇒ ↑)
                                                     ≡⟨ ≡-cong λ²⇓_ (≡-cong₂ _/⊢_ (byDef {x = ⌜ e ⌝})
                                                                                  (⌜↑⌝-lemma σ₂ (_↑ {σ = σ₁} ρ))) ⟩
    λ²⇓ (⌜ e ⌝ /⊢ ⌜ _↑ {σ = σ₂} (_↑ {σ = σ₁} ρ) ⌝⇒)
                                                     ≡⟨ ≡-cong λ²⇓_ (⌜/⊢⌝-lemma e (ρ ↑ ↑)) ⟩
    λ²⇓ ⌜ e /⊢ ρ ↑ ↑ ⌝
                                                     ∎
  ⌜/⊢⇓⌝-lemma z⇓ ρ = byDef
  ⌜/⊢⇓⌝-lemma s⇓ ρ = byDef
  ⌜/⊢⇓⌝-lemma (s₁⇓ x) ρ = ≡-cong s₁⇓ (⌜/∋⌝-lemma x ρ)
  ⌜/⊢⇓⌝-lemma (natrec⇓ {σ = σ} {n = n} x₁ e₂) ρ =
                                                    begin
    f (⌜ e₂ ⌝ /⊢ ⌜ ρ ⌝⇒ ↑ ↑)
                                                    ≡⟨ ≡-cong f (≡-cong₂ _/⊢_ (byDef {x = ⌜ e₂ ⌝})
                                                                              (≡-cong _↑ (⌜↑⌝-lemma σ₁ ρ))) ⟩
    f (⌜ e₂ ⌝ /⊢ ⌜ _↑ {σ = σ₁} ρ ⌝⇒ ↑)
                                                    ≡⟨ ≡-cong f (≡-cong₂ _/⊢_ (byDef {x = ⌜ e₂ ⌝})
                                                                              (⌜↑⌝-lemma σ₂ (_↑ {σ = σ₁} ρ))) ⟩
    f (⌜ e₂ ⌝ /⊢ ⌜ _↑ {σ = σ₂} (_↑ {σ = σ₁} ρ) ⌝⇒)
                                                    ≡⟨ ≡-cong₂ natrec⇓ (⌜/∋⌝-lemma x₁ ρ)
                                                                       (⌜/⊢⌝-lemma e₂ (ρ ↑ ↑)) ⟩
    natrec⇓ ⌜ x₁ /∋ ρ ⌝∋ ⌜ e₂ /⊢ ρ ↑ ↑ ⌝
                                                    ∎
    where
    σ₁ = Nat
    σ₂ = Thunk (suc n) σ
    f  = natrec⇓ (⌜ x₁ ⌝∋ /∋ ⌜ ρ ⌝⇒)
  ⌜/⊢⇓⌝-lemma (x₁ ,⇓ x₂) ρ = ≡-cong₂ _,⇓_ (⌜/∋⌝-lemma x₁ ρ) (⌜/∋⌝-lemma x₂ ρ)
  ⌜/⊢⇓⌝-lemma (uncurry⇓ {σ₁ = σ₁} {σ₂ = σ₂} e) ρ =
                                                          begin
    uncurry⇓ (⌜ e ⌝ /⊢ ⌜ ρ ⌝⇒ ↑ ↑)
                                                          ≡⟨ ≡-cong uncurry⇓ (≡-cong₂ _/⊢_ (byDef {x = ⌜ e ⌝})
                                                                                           (≡-cong _↑ (⌜↑⌝-lemma σ₁ ρ))) ⟩
    uncurry⇓ (⌜ e ⌝ /⊢ ⌜ _↑ {σ = σ₁} ρ ⌝⇒ ↑)
                                                          ≡⟨ ≡-cong uncurry⇓ (≡-cong₂ _/⊢_ (byDef {x = ⌜ e ⌝})
                                                                                           (⌜↑⌝-lemma σ₂ (_↑ {σ = σ₁} ρ))) ⟩
    uncurry⇓ (⌜ e ⌝ /⊢ ⌜ _↑ {σ = σ₂} (_↑ {σ = σ₁} ρ) ⌝⇒)
                                                          ≡⟨ ≡-cong uncurry⇓ (⌜/⊢⌝-lemma e (ρ ↑ ↑)) ⟩
    uncurry⇓ ⌜ e /⊢ ρ ↑ ↑ ⌝
                                                          ∎
  ⌜/⊢⇓⌝-lemma (return⋆⇓ e) ρ = ⌜/⊢⇓⌝-lemma e ρ

  ⌜return⋆⌝-lemma :  forall {Γ σ} (e : Γ ⊢ σ) n
                  -> ⌜ e ⌝ ≡ ⌜ return⋆ {n = n} e ⌝
  ⌜return⋆⌝-lemma e zero    = byDef
  ⌜return⋆⌝-lemma e (suc n) = ⌜return⋆⌝-lemma e n

  ⌜valToTm⌝-lemma :  forall {Γ σ} (v : Γ ⊢⇓ σ)
                  -> valToTm ⌜ v ⌝⇓ ≡ ⌜ valToTm v ⌝
  ⌜valToTm⌝-lemma (λ⇓ e)               = byDef
  ⌜valToTm⌝-lemma (λ²⇓ e)              = byDef
  ⌜valToTm⌝-lemma z⇓                   = byDef
  ⌜valToTm⌝-lemma s⇓                   = byDef
  ⌜valToTm⌝-lemma (s₁⇓ x)              = byDef
  ⌜valToTm⌝-lemma (natrec⇓ x₁ e₂)      = byDef
  ⌜valToTm⌝-lemma (x ,⇓ y)             = byDef
  ⌜valToTm⌝-lemma (uncurry⇓ e)         = byDef
  ⌜valToTm⌝-lemma (return⋆⇓ {n = n} v) =
                                     begin
    valToTm ⌜ v ⌝⇓
                                     ≡⟨ ⌜valToTm⌝-lemma v ⟩
    ⌜ valToTm v ⌝
                                     ≡⟨ ⌜return⋆⌝-lemma (valToTm v) n ⟩
    ⌜ return⋆ {n = n} (valToTm v) ⌝
                                     ∎

  ⌜drop-return⋆⇓⌝-lemma :  forall {Γ τ} ns (v : Γ ⊢⇓ Thunks ns τ)
                        -> ≡-subst (_⊢⇓_ _) (Thunks-lemma ns) ⌜ v ⌝⇓ ≡
                           ⌜ drop-return⋆⇓ ns v ⌝⇓
  ⌜drop-return⋆⇓⌝-lemma []       v            = byDef
  ⌜drop-return⋆⇓⌝-lemma (n ∷ ns) (return⋆⇓ v) =
    ⌜drop-return⋆⇓⌝-lemma ns v

  ⌜gwait⇓⌝-lemma :  forall {Γ σ n₁} C n₂ (v : Γ ⊢⇓ C [ Thunk n₁ σ ])
                 -> ≡-subst (_⊢⇓_ _) (erase-lemma C) ⌜ v ⌝⇓ ≡
                    ⌜ gwait⇓ C n₂ v ⌝⇓
  ⌜gwait⇓⌝-lemma ●          n₂ (return⋆⇓ v) = byDef
  ⌜gwait⇓⌝-lemma (const● σ) n₂ v            = byDef
  ⌜gwait⇓⌝-lemma {σ = τ} {n₁ = n₁} (C ●× σ) n₂ (x ,⇓ y)
    with ⌜ C [ Thunk (n₂ + n₁) τ ] ⌝⋆
       | erase-lemma {m = n₁} {n = n₂ + n₁} {τ = τ} C
  ... | ._ | ≡-refl = ≡-refl
  ⌜gwait⇓⌝-lemma {σ = τ} {n₁ = n₁} (σ ×● C) n₂ (x ,⇓ y)
    with ⌜ C [ Thunk (n₂ + n₁) τ ] ⌝⋆
       | erase-lemma {m = n₁} {n = n₂ + n₁} {τ = τ} C
  ... | ._ | ≡-refl = ≡-refl
  ⌜gwait⇓⌝-lemma (Thunk● n C) n₂ (return⋆⇓ v) = ⌜gwait⇓⌝-lemma C n₂ v
