------------------------------------------------------------------------
-- Lemmas about GPay
------------------------------------------------------------------------

module OperationalSemantics.GPay.Lemmas where

open import Kind
open import TypeSystem
open import Erasure
open import Substitution
open VarSubst
open import OperationalSemantics.Binding
open Heap↦
open Heap↦Erasure
open import Value
open import OperationalSemantics.GPay
open import Erasure.Lemmas
open import Erasure.Solver
open import Substitution.Lemmas
open import PayCtxt
open import TypeTriple

open import Data.Nat
open import Data.Nat.Properties
open import Data.List using (_∷_)
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning

------------------------------------------------------------------------
-- Boring cast lemmas

cast-∘ˢ-commute
  :  forall {Γ Δ¹ Δ² : Ctxt simple} (ρ : Γ ⇒ Δ¹) (eq : Δ¹ ≡ Δ²)
  -> ≡-subst (_⇒_ Γ) eq (ρ ∘ˢ idˢ) ≡ ρ ∘ˢ ≡-subst (_⇒_ Δ¹) eq idˢ
cast-∘ˢ-commute ρ ≡-refl = ≡-refl

cast-/∋-commute
  :  forall {Γ¹ Γ² : Ctxt simple} {σ} (x : Γ¹ ∋ σ) (eq : Γ¹ ≡ Γ²)
  -> ≡-subst (\Γ -> Γ ∋ σ) eq x /∋ idˢ ≡ x /∋ ≡-subst (_⇒_ Γ¹) eq idˢ
cast-/∋-commute x ≡-refl = ≡-refl

cast-/↦-commute
  :  forall {Γ¹ Γ² : Ctxt simple} {σ} (b : Γ¹ ↦ σ) (eq : Γ¹ ≡ Γ²)
  -> ≡-subst (\Γ -> Γ ↦ σ) eq b /↦ idˢ ≡ b /↦ ≡-subst (_⇒_ Γ¹) eq idˢ
cast-/↦-commute b ≡-refl = ≡-refl

cast-lemma₁
  :  forall {Γ Γ' : Ctxt simple} {σ₁ σ₁' σ₂} x₁ x₂
            (eq₁ : Γ ≡ Γ') (eq₂ : σ₁ ≡ σ₁')
  -> ≡-subst (\Δ -> Δ ⊢⇓ σ₁' × σ₂) eq₁
       (≡-subst (_⊢⇓_ Γ) (≡-cong (\τ -> τ × σ₂) eq₂) (x₁ ,⇓ x₂)) ≡
     ≡-subst (\Δ -> Δ ∋ σ₁') eq₁ (≡-subst (_∋_ Γ) eq₂ x₁) ,⇓
     ≡-subst (\Δ -> Δ ∋ σ₂) eq₁ x₂
cast-lemma₁ x₁ x₂ ≡-refl ≡-refl = ≡-refl

cast-lemma₂
  :  forall {Γ Γ' : Ctxt simple} {σ₁ σ₂ σ₂'} x₁ x₂
            (eq₁ : Γ ≡ Γ') (eq₂ : σ₂ ≡ σ₂')
  -> ≡-subst (\Δ -> Δ ⊢⇓ σ₁ × σ₂') eq₁
       (≡-subst (_⊢⇓_ Γ) (≡-cong (\τ -> σ₁ × τ) eq₂) (x₁ ,⇓ x₂)) ≡
     ≡-subst (\Δ -> Δ ∋ σ₁) eq₁ x₁ ,⇓
     ≡-subst (\Δ -> Δ ∋ σ₂') eq₁ (≡-subst (_∋_ Γ) eq₂ x₂)
cast-lemma₂ x₁ x₂ ≡-refl ≡-refl = ≡-refl

cast-lemma₃
  :  forall {Γ : Ctxt simple} {σ τ σ₁ σ₂ σ₃ σ₄} (t : Γ ⊢ σ)
            (eq₁ : σ₁ ≡ τ) (eq₂ : σ ≡ σ₁)
            (eq₃ : σ₄ ≡ τ) (eq₄ : σ₃ ≡ σ₄)
            (eq₅ : σ₂ ≡ σ₃) (eq₆ : σ ≡ σ₂)
  -> ≡-subst (_↦_ Γ) eq₁ (↦ ≡-subst (_⊢_ Γ) eq₂ t) ≡
     ↦ ≡-subst (_⊢_ Γ) eq₃ (≡-subst (_⊢_ Γ) eq₄
         (≡-subst (_⊢_ Γ) eq₅ (≡-subst (_⊢_ Γ) eq₆ t)))
cast-lemma₃ t ≡-refl ≡-refl ≡-refl ≡-refl ≡-refl ≡-refl = ≡-refl

cast-lemma₄
  :  forall {k} {Γ : Ctxt k} {σ₁ σ₂} {Σ b₂} (b₁ : Γ ↦ σ₁)
            (eq : σ₁ ≡ σ₂)
  -> ≡-subst (_↦_ Γ) eq b₁ ≡ b₂
  -> ≡-subst Heap (≡-cong (_▻_ Γ) eq) (Σ ▷ʰ b₁) ≡ Σ ▷ʰ b₂
cast-lemma₄ _ ≡-refl ≡-refl = ≡-refl

cast-lemma₅
  :  forall {k} {Γ : Ctxt k} {σ₁ σ₂} (eq : σ₁ ≡ σ₂)
  -> ≡-subst (\Γ' -> Γ' ∋ σ₂) (≡-cong (_▻_ Γ) eq)
             (≡-subst (_∋_ _) eq vz) ≡ vz
cast-lemma₅ ≡-refl = ≡-refl

cast-lemma₆
  :  forall {Γ₁ Γ₂ σ₁ σ₂ τ} (x : Γ₁ ∋ σ₁) {y : Γ₂ ∋ σ₂}
            (eq₁ : ⌜ σ₁ ⌝⋆ ≡ σ₂) (eq₂ : ⌜ Γ₁ ⌝▻ ≡ Γ₂)
  -> ≡-subst (\Γ -> Γ ∋ σ₂) eq₂ (≡-subst (_∋_ _) eq₁ ⌜ x ⌝∋) ≡ y
  -> ≡-subst (\Γ -> Γ ∋ σ₂) (≡-cong (\Γ -> Γ ▻ τ) eq₂)
             (≡-subst (_∋_ _) eq₁ (vs ⌜ x ⌝∋)) ≡ vs y
cast-lemma₆ _ ≡-refl ≡-refl ≡-refl = ≡-refl

cast-lemma₇
  :  forall {k} {Γ : Ctxt k} {σ₁ σ₂ τ}
            (eq₁ : σ₁ ≡ σ₂) (eq₂ : τ ≡ σ₁) (eq₃ : σ₂ ≡ τ)
  -> ≡-subst (_⇒_ _) (≡-cong (_▻_ _) eq₁) (wk {Γ = Γ} ▷ vz {σ = σ₁}) ≡
     wk ▷ ≡-subst (_∋_ _) eq₂ (≡-subst (_∋_ _) eq₃ (vz {σ = σ₂}))
cast-lemma₇ ≡-refl ≡-refl ≡-refl = ≡-refl

cast-lemma₈
  :  forall {k} {Γ Γ' : Ctxt k} {σ} (Σ : Heap Γ) (b : Γ ↦ σ)
                (eq : Γ ≡ Γ')
  -> ≡-subst Heap (≡-cong (\Γ -> Γ ▻ σ) eq) (Σ ▷ʰ b) ≡
     ≡-subst Heap eq Σ ▷ʰ ≡-subst (\Γ -> Γ ↦ σ) eq b
cast-lemma₈ _ _ ≡-refl = ≡-refl

cast-lemma₉
  :  forall {k} {Γ Γ' : Ctxt k} {σ} (eq : Γ ≡ Γ')
  -> ≡-subst (_⇒_ _) (≡-cong (\Γ -> Γ ▻ σ) eq) (idˢ {Γ = Γ ▻ σ}) ≡
     ≡-subst (_⇒_ _) eq idˢ ↑
cast-lemma₉ ≡-refl = ≡-refl

cast-lemma₁₀
  :  forall {Γ₁ Γ₂ n n' σ₁ σ₂ σ₂' σ₃ σ₃' σ₄} C C' m m'
     (x₁ : Γ₁ ∋ C [ Thunk n σ₁ ]) (x₂ : Γ₂ ∋ σ₃)
     (Δ-eq  : ⌜ Γ₁ ⌝▻ ≡ ⌜ Γ₂ ⌝▻)
     (Δ-eq' : ⌜ Γ₁ ⌝▻ ≡ ⌜ Γ₂ ⌝▻)
     (erase-eq  : ⌜ σ₂  ⌝⋆ ≡ ⌜ σ₃  ⌝⋆)
     (erase-eq' : ⌜ σ₂' ⌝⋆ ≡ ⌜ σ₃' ⌝⋆)
     (●-eq  : C [ Thunk n σ₁ ] ≡ σ₂)
     (●-eq' : C [ Thunk (m + n) σ₁ ] ≡ σ₂')
     (eq₁ : C' [ Thunk (m' + n') σ₄ ] ≡ σ₃')
     (eq₂ : σ₃ ≡ C' [ Thunk n' σ₄ ])
  -> ≡-subst (\Γ -> Γ ∋ ⌜ σ₃ ⌝⋆) Δ-eq
       (≡-subst (_∋_ ⌜ Γ₁ ⌝▻) erase-eq
          ⌜ ≡-subst (_∋_ Γ₁) ●-eq x₁ ⌝∋) ≡
     ⌜ x₂ ⌝∋
  -> ≡-subst (\Γ -> Γ ∋ ⌜ σ₃' ⌝⋆) Δ-eq'
       (≡-subst (_∋_ ⌜ Γ₁ ⌝▻) erase-eq'
          ⌜ ≡-subst (_∋_ Γ₁) ●-eq' (gwait∋ C m x₁) ⌝∋) ≡
     ⌜ ≡-subst (_∋_ Γ₂) eq₁ (gwait∋ C' m' (≡-subst (_∋_ Γ₂) eq₂ x₂)) ⌝∋
cast-lemma₁₀ {n = n} {n' = n'} {σ₁ = σ₁} {σ₄ = σ₄}
             C C' m m' x₁ x₂ Δ-eq Δ-eq' erase-eq erase-eq'
             ≡-refl ≡-refl ≡-refl ≡-refl hyp
  with ⌜ x₂ ⌝∋ | hyp | ⌜ C' [ Thunk (m' + n') σ₄ ] ⌝⋆
     | erase-lemma {m = n'} {n = m' + n'} {τ = σ₄} C'
     | ⌜ C [ Thunk (m + n) σ₁ ] ⌝⋆
     | erase-lemma {m = n}  {n = m + n}   {τ = σ₁} C
... | ._ | ≡-refl | ._ | ≡-refl | ._ | ≡-refl =
  helper Δ-eq' Δ-eq erase-eq' erase-eq
  where
  helper :  forall {Γ₁ Γ₂ : Ctxt simple} {σ₁ σ₂ x}
            (eq₁ eq₂ : Γ₁ ≡ Γ₂) (eq₃ eq₄ : σ₁ ≡ σ₂)
         -> ≡-subst (\Γ -> Γ ∋ σ₂) eq₁ (≡-subst (_∋_ Γ₁) eq₃ x) ≡
            ≡-subst (\Γ -> Γ ∋ σ₂) eq₂ (≡-subst (_∋_ Γ₁) eq₄ x)
  helper ≡-refl ≡-refl ≡-refl ≡-refl = ≡-refl

------------------------------------------------------------------------
-- Another lemma

≜-⁻¹-lemma :  forall C {C₁ n₁ σ₁ C₂ n₂ σ₂}
           -> ⟨ C₁ , n₁ , σ₁ ⟩ ≜ ⟨ C₂ , n₂ , σ₂ ⟩
           -> C ⁻¹∘ C₁ [ Thunk n₁ σ₁ ] ≡ C ⁻¹∘ C₂ [ Thunk n₂ σ₂ ]
≜-⁻¹-lemma C {C₁} {n₁} {σ₁} {C₂} {n₂} {σ₂} eq = begin
  C ⁻¹∘ C₁ [ Thunk n₁ σ₁ ]
                              ≡⟨ ≡-sym (⁻¹[]-lemma C C₁) ⟩
  C ⁻¹[ C₁ [ Thunk n₁ σ₁ ] ]
                              ≡⟨ ≡-cong (_⁻¹[_] C) (sound eq) ⟩
  C ⁻¹[ C₂ [ Thunk n₂ σ₂ ] ]
                              ≡⟨ ⁻¹[]-lemma C C₂ ⟩
  C ⁻¹∘ C₂ [ Thunk n₂ σ₂ ]
                              ∎

------------------------------------------------------------------------
-- Type used to state correctness of gpayResult

record GPaySound (_•_ : TmLike)
                 (⌜_⌝• : TmEraser _•_)
                 {Γ : Ctxt thunked} {C : PayCtxt}
                 {n m : ℕ} {σ : Ty thunked}
                 (result : GPay _•_ Γ C n m σ)
                 (Σ₁ : Heap Γ) (v₁ : Γ • C [ Thunk n σ ])
                 : Set where
  field
    Δ-eq  : ⌜ Γ ⌝▻                        ≡ ⌜ GPay.Δ  result ⌝▻
    ρ-eq  : ≡-subst (_⇒_ ⌜ Γ ⌝▻) Δ-eq idˢ ≡ ⌜ GPay.ρ  result ⌝⇒
    Σ₂-eq : ≡-subst Heap Δ-eq ⌜ Σ₁ ⌝ʰ     ≡ ⌜ GPay.Σ₂ result ⌝ʰ
    v₂-eq : ≡-subst (\Γ -> Γ • ⌜ C [ Thunk (n ∸ m) σ ] ⌝⋆) Δ-eq
              (≡-subst (_•_ ⌜ Γ ⌝▻) (erase-lemma C) ⌜ v₁ ⌝•) ≡
            ⌜ GPay.v₂ result ⌝•

-- Another cast lemma (proof _irr_elevance).

irr  :  forall {_•_ : TmLike} {⌜_⌝• : TmEraser _•_}
               {Γ C n m σ σ'}
               {r : GPay _•_ Γ C n m σ}
               {Σ₁ v₁}
        (eq eq' : σ' ≡ C [ Thunk n σ ])
     -> GPaySound _•_ ⌜_⌝• r Σ₁ (≡-subst (_•_ Γ) eq  v₁)
     -> GPaySound _•_ ⌜_⌝• r Σ₁ (≡-subst (_•_ Γ) eq' v₁)
irr ≡-refl ≡-refl rs = rs

-- Some lemmas which make it easier to make use of the equalities
-- above.

∘-ρ-lemma
  :  forall {_•_ : TmLike} {⌜_⌝• : TmEraser _•_}
            {Γ C n m σ}
            {r : GPay _•_ Γ C n m σ}
            {Σ₁ v₁}
  -> (rs : GPaySound _•_ ⌜_⌝• r Σ₁ v₁)
  -> forall {X} (ρ : X ⇒ Γ)
  -> ≡-subst (_⇒_ _) (GPaySound.Δ-eq rs) ⌜ ρ ⌝⇒ ≡ ⌜ ρ ∘ˢ GPay.ρ r ⌝⇒
∘-ρ-lemma {r = r} rs ρ₁ = begin
  cast₁ ⌜ ρ₁ ⌝⇒
                          ≡⟨ ≡-cong cast₁ (≡-sym (∘-id-lemma ⌜ ρ₁ ⌝⇒)) ⟩
  cast₁ (⌜ ρ₁ ⌝⇒ ∘ˢ idˢ)
                          ≡⟨ cast-∘ˢ-commute ⌜ ρ₁ ⌝⇒ Δ-eq ⟩
  ⌜ ρ₁ ⌝⇒ ∘ˢ cast₂ idˢ
                          ≡⟨ ≡-cong (_∘ˢ_ ⌜ ρ₁ ⌝⇒) (GPaySound.ρ-eq rs) ⟩
  ⌜ ρ₁ ⌝⇒ ∘ˢ ⌜ ρ₂ ⌝⇒
                          ≡⟨ prove (|⌜ lift⇒ ρ₁ ⌝⇒| |∘ˢ| |⌜ lift⇒ ρ₂ ⌝⇒|)
                                   |⌜ lift⇒ ρ₁ |∘ˢ| lift⇒ ρ₂ ⌝⇒|
                                   ≡-refl ⟩
  ⌜ ρ₁ ∘ˢ ρ₂ ⌝⇒
                          ∎
  where
  Δ-eq  = GPaySound.Δ-eq rs
  ρ₂    = GPay.ρ r
  cast₁ = ≡-subst (_⇒_ _) Δ-eq
  cast₂ = ≡-subst (_⇒_ _) Δ-eq

/∋-ρ-lemma
  :  forall {_•_ : TmLike} {⌜_⌝• : TmEraser _•_}
            {Γ C n m σ}
            {r : GPay _•_ Γ C n m σ}
            {Σ₁ v₁}
  -> (rs : GPaySound _•_ ⌜_⌝• r Σ₁ v₁)
  -> forall {τ} (x : Γ ∋ τ)
  -> ≡-subst (\Γ -> Γ ∋ ⌜ τ ⌝⋆) (GPaySound.Δ-eq rs) ⌜ x ⌝∋ ≡
     ⌜ x /∋ GPay.ρ r ⌝∋
/∋-ρ-lemma {r = r} rs {τ = τ} x = begin
  cast₁ ⌜ x ⌝∋
                           ≡⟨ ≡-sym (/∋-id-lemma (cast₁ ⌜ x ⌝∋)) ⟩
  cast₁ ⌜ x ⌝∋ /∋ idˢ
                           ≡⟨ cast-/∋-commute ⌜ x ⌝∋ Δ-eq ⟩
  ⌜ x ⌝∋ /∋ cast₂ idˢ
                           ≡⟨ ≡-cong (_/∋_ ⌜ x ⌝∋) (GPaySound.ρ-eq rs) ⟩
  ⌜ x ⌝∋ /∋ ⌜ GPay.ρ r ⌝⇒
                           ≡⟨ prove (|⌜ lift x ⌝∋| |/∋| |⌜ lift⇒ (GPay.ρ r) ⌝⇒|)
                                    |⌜ lift x |/∋| lift⇒ (GPay.ρ r) ⌝∋|
                                    ≡-refl ⟩
  ⌜ x /∋ GPay.ρ r ⌝∋
                           ∎
  where
  Δ-eq  = GPaySound.Δ-eq rs
  ρ₂    = GPay.ρ r
  cast₁ = ≡-subst (\Γ -> Γ ∋ ⌜ τ ⌝⋆) Δ-eq
  cast₂ = ≡-subst (_⇒_ _) Δ-eq

/↦-ρ-lemma
  :  forall {_•_ : TmLike} {⌜_⌝• : TmEraser _•_}
            {Γ C n m σ}
            {r : GPay _•_ Γ C n m σ}
            {Σ₁ v₁}
  -> (rs : GPaySound _•_ ⌜_⌝• r Σ₁ v₁)
  -> forall {τ} (b : Γ ↦ τ)
  -> ≡-subst (\Γ -> Γ ↦ ⌜ τ ⌝⋆) (GPaySound.Δ-eq rs) ⌜ b ⌝↦ ≡
     ⌜ b /↦ GPay.ρ r ⌝↦
/↦-ρ-lemma {r = r} rs {τ = τ} b = begin
  cast₁ ⌜ b ⌝↦
                           ≡⟨ ≡-sym (/↦-id-lemma (cast₁ ⌜ b ⌝↦)) ⟩
  cast₁ ⌜ b ⌝↦ /↦ idˢ
                           ≡⟨ cast-/↦-commute ⌜ b ⌝↦ Δ-eq ⟩
  ⌜ b ⌝↦ /↦ cast₂ idˢ
                           ≡⟨ ≡-cong (_/↦_ ⌜ b ⌝↦) (GPaySound.ρ-eq rs) ⟩
  ⌜ b ⌝↦ /↦ ⌜ GPay.ρ r ⌝⇒
                           ≡⟨ prove (|⌜ lift b ⌝↦| |/↦| |⌜ lift⇒ (GPay.ρ r) ⌝⇒|)
                                    |⌜ lift b |/↦| lift⇒ (GPay.ρ r) ⌝↦|
                                    ≡-refl ⟩
  ⌜ b /↦ GPay.ρ r ⌝↦
                           ∎
  where
  Δ-eq  = GPaySound.Δ-eq rs
  ρ₂    = GPay.ρ r
  cast₁ = ≡-subst (\Γ -> Γ ↦ ⌜ τ ⌝⋆) Δ-eq
  cast₂ = ≡-subst (_⇒_ _) Δ-eq

------------------------------------------------------------------------
-- The various GPay functions erase to nothing

gpay↦-sound :  forall {Γ σ n} C m (b : Γ ↦ C [ Thunk n σ ])
            -> ≡-subst (_↦_ ⌜ Γ ⌝▻) (erase-lemma C) ⌜ b ⌝↦ ≡
               ⌜ gpay↦ C m b ⌝↦
gpay↦-sound {Γ = Γ} C m (↦ⁿˢ ns e) = begin
  cast₃ (↦ cast₄' ⌜ e ⌝)
                                                       ≡⟨ cast-lemma₃ ⌜ e ⌝
                                                            lem₅ lem₃' lem₃
                                                            (≡-cong ⌜_⌝⋆ lem₂) lem₄
                                                            (≡-cong ⌜_⌝⋆ lem₁)  ⟩
  ↦ cast₄ (cast₅ (cast₆ (cast₇ ⌜ e ⌝)))
                                                       ≡⟨ lemma ⟩
  ↦ cast₄ ⌜ cast₁ (gpay (Thunks● ns C) m (cast₂ e)) ⌝
                                                       ∎
  where
  lem₁   = Thunks●-lemma ns C
  lem₂   = ≡-sym (Thunks●-lemma (m ∷ ns) C)
  lem₃   = Thunks-lemma ns
  lem₃'  = Thunks-lemma ns
  lem₄   = erase-lemma (Thunks● ns C)
  lem₅   = erase-lemma C
  cast₁  = ≡-subst (_⊢_ Γ) lem₂
  cast₂  = ≡-subst (_⊢_ Γ) lem₁
  cast₃  = ≡-subst (_↦_ ⌜ Γ ⌝▻) lem₅
  cast₄  = ≡-subst (_⊢_ ⌜ Γ ⌝▻) lem₃
  cast₄' = ≡-subst (_⊢_ ⌜ Γ ⌝▻) lem₃'
  cast₅  = ≡-subst (_⊢_ ⌜ Γ ⌝▻) (≡-cong ⌜_⌝⋆ lem₂)
  cast₆  = ≡-subst (_⊢_ ⌜ Γ ⌝▻) lem₄
  cast₇  = ≡-subst (_⊢_ ⌜ Γ ⌝▻) (≡-cong ⌜_⌝⋆ lem₁)

  lemma = prove
    (|↦| |≡-subst⊢| lem₃
         (|≡-subst⊢| (≡-cong ⌜_⌝⋆ lem₂)
                     (|≡-subst⊢| (erase-lemma (Thunks● ns C))
                     (|≡-subst⊢| (≡-cong ⌜_⌝⋆ lem₁) |⌜ lift e ⌝|))))
    (|↦| |≡-subst⊢| lem₃
         |⌜ |≡-subst⊢| lem₂
                       (|gpay| (Thunks● ns C) m
                               (|≡-subst⊢| lem₁ (lift e))) ⌝|)
    ≡-refl

mutual

  gpayResult∋-sound :  forall C m {Γ τ n σ} (Σ₁ : Heap Γ)
                              (x₁ : Γ ∋ τ) (eq : τ ≡ C [ Thunk n σ ])
                    -> GPaySound _∋_ ⌜_⌝∋ (gpayResult∋ C m Σ₁ x₁ eq)
                                          Σ₁ (≡-subst (_∋_ _) eq x₁)
  gpayResult∋-sound C₂ m Σ₁ (gwait∋ C₁ n x₁) eq =
    irr eq' eq (gpayResult∋-gwait-sound ● m n _ _ Σ₁ x₁
                  (complete ⟨ C₁ , _ , _ ⟩ ⟨ C₂ , _ , _ ⟩ eq))
    where eq' = ≜-⁻¹-lemma ● (complete ⟨ C₁ , _ , _ ⟩ ⟨ C₂ , _ , _ ⟩ eq)
  gpayResult∋-sound C m {n = n} {σ = σ} (Σ₁ ▷ʰ b) (vz {Γ = Γ}) ≡-refl =
    record
      { Δ-eq  = Δ-eq
      ; ρ-eq  = begin
          cast₀ idˢ
                                                ≡⟨ byDef ⟩
          cast₀ (wk ▷ vz)
                                                ≡⟨ cast-lemma₇ (erase-lemma C)
                                                               (≡-cong ⌜_⌝⋆ lem₁)
                                                               lem₂ ⟩
          wk ▷ cast₂ (cast₃ vz)
                                                ≡⟨ lemma ⟩
          ⌜ wk ▷ cast₁ (gwait∋ C (m ⊓ n) x) ⌝⇒
                                                ∎
      ; Σ₂-eq = cast-lemma₄ ⌜ b ⌝↦ (erase-lemma C) (gpay↦-sound C m b)
      ; v₂-eq = cast-lemma₅ (erase-lemma C)
      }
    where
    Δ-eq  = ≡-cong (_▻_ ⌜ Γ ⌝▻) (erase-lemma C)
    lem₁  = ≡-cong (\i -> C [ Thunk i σ ]) (m⊓n+n∸m≡n m n)
    lem₂  = erase-lemma {m = n ∸ m} {n = (m ⊓ n) + (n ∸ m)} C
    cast₀ = ≡-subst (_⇒_ _) Δ-eq
    cast₁ = ≡-subst (_∋_ _) lem₁
    cast₂ = ≡-subst (_∋_ _) (≡-cong ⌜_⌝⋆ lem₁)
    cast₃ = ≡-subst (_∋_ _) lem₂
    x     = vz {Γ = Γ} {σ = C [ Thunk (n ∸ m) σ ]}

    lemma = prove
      (|wk| |▷| |≡-subst∋| (≡-cong ⌜_⌝⋆ lem₁)
                   (|≡-subst∋| lem₂ |⌜ lift x ⌝∋|))
      |⌜ |wk| |▷| |≡-subst∋| lem₁ (|gwait∋| C (m ⊓ n) (lift x)) ⌝⇒|
      ≡-refl

  gpayResult∋-sound C m (Σ₁ ▷ʰ b) (vs {Γ = Γ} {σ = σ} x₁) ≡-refl = record
    { Δ-eq  = Δ-eq
    ; ρ-eq  = begin
        ≡-subst (_⇒_ _) Δ-eq idˢ
                                    ≡⟨ cast-lemma₉ lem ⟩
        ≡-subst (_⇒_ _) lem idˢ ↑
                                    ≡⟨ ≡-cong _↑ (GPaySound.ρ-eq rs) ⟩
        ⌜ GPay.ρ r ⌝⇒ ↑
                                    ≡⟨ prove (|⌜ lift⇒ (GPay.ρ r) ⌝⇒| |↑|)
                                             |⌜ _|↑| {σ = σ} (lift⇒ (GPay.ρ r)) ⌝⇒|
                                             ≡-refl ⟩
        ⌜ _↑ {σ = σ} (GPay.ρ r) ⌝⇒
                                    ∎
    ; Σ₂-eq = begin
        ≡-subst Heap Δ-eq (⌜ Σ₁ ⌝ʰ ▷ʰ ⌜ b ⌝↦)
                                                ≡⟨ cast-lemma₈ ⌜ Σ₁ ⌝ʰ ⌜ b ⌝↦ lem ⟩
        ≡-subst Heap lem ⌜ Σ₁ ⌝ʰ ▷ʰ
        ≡-subst (\Γ -> Γ ↦ ⌜ σ ⌝⋆) lem ⌜ b ⌝↦
                                                ≡⟨ ≡-cong₂ _▷ʰ_ (GPaySound.Σ₂-eq rs)
                                                                (/↦-ρ-lemma rs b) ⟩
        ⌜ GPay.Σ₂ r ⌝ʰ ▷ʰ ⌜ b /↦ GPay.ρ r ⌝↦
                                                ∎
    ; v₂-eq = cast-lemma₆ x₁ (erase-lemma C) (GPaySound.Δ-eq rs)
                             (GPaySound.v₂-eq rs)
    }
    where
    r    = gpayResult∋ C m Σ₁ x₁ ≡-refl
    rs   = gpayResult∋-sound C m Σ₁ x₁ ≡-refl
    lem  = GPaySound.Δ-eq rs
    Δ-eq = ≡-cong (\Γ -> Γ ▻ ⌜ σ ⌝⋆) lem

  gpayResult∋-gwait-sound
    :  forall C {C₁ C₂} (m n n₁ n₂ : ℕ) {Γ σ₁ σ₂} (Σ₁ : Heap Γ)
    -> (x₁ : Γ ∋ C ⁻¹∘ C₁ [ Thunk n₁ σ₁ ])
    -> (eq : ⟨ C₁ , n + n₁ , σ₁ ⟩ ≜ ⟨ C₂ , n₂ , σ₂ ⟩)
    -> GPaySound _∋_ ⌜_⌝∋ (gpayResult∋-gwait C m n n₁ n₂ Σ₁ x₁ eq) Σ₁
                 (≡-subst (_∋_ Γ) (≜-⁻¹-lemma C eq)
                          (gwait∋ (C ⁻¹∘ C₁) n x₁))
  gpayResult∋-gwait-sound C m n n₁ n₂ {σ₂ = σ₂} Σ₁ x₁ (≜c- {C = C₂}) =
    record
      { Δ-eq  = GPaySound.Δ-eq  rs
      ; ρ-eq  = GPaySound.ρ-eq  rs
      ; Σ₂-eq = GPaySound.Σ₂-eq rs
      ; v₂-eq = cast-lemma₁₀
                  (C ⁻¹∘ const● (C₂ [ Thunk n₂ σ₂ ])) C'
                  n zero x₁ (GPay.v₂ r)
                  (GPaySound.Δ-eq rs)
                  (GPaySound.Δ-eq rs)
                  (erase-lemma C')
                  (erase-lemma C')
                  (⁻¹-const●-lemma C C₂)
                  (≜-⁻¹-lemma C ≜c-)
                  ≡-refl
                  ≡-refl
                  (GPaySound.v₂-eq rs)
      }
    where
    C'  = C ⁻¹∘ C₂
    lem = ⁻¹-const●-lemma C C₂
    r   = gpayResult∋       C' m Σ₁ x₁ lem
    rs  = gpayResult∋-sound C' m Σ₁ x₁ lem

  gpayResult∋-gwait-sound C m n n₁ n₂ Σ₁ x₁ (≜-c {C = C₁} {σ₁ = σ₁}) =
    record
      { Δ-eq  = GPaySound.Δ-eq  rs
      ; ρ-eq  = GPaySound.ρ-eq  rs
      ; Σ₂-eq = GPaySound.Σ₂-eq rs
      ; v₂-eq = cast-lemma₁₀
                  (C ⁻¹∘ C₁) (C ⁻¹∘ C₁)
                  n n x₁ (GPay.v₂ r)
                  (GPaySound.Δ-eq rs)
                  (GPaySound.Δ-eq rs)
                  (erase-lemma C'')
                  (erase-lemma (C ⁻¹∘ const● (C₁ [ Thunk (n + n₁) σ₁ ])))
                  (≡-sym (⁻¹-const●-lemma C C₁))
                  (≜-⁻¹-lemma C ≜-c)
                  (≡-sym (⁻¹-const●-lemma C C₁))
                  (⁻¹-const●-lemma C C₁)
                  (GPaySound.v₂-eq rs)
      }
    where
    C'' = C ⁻¹∘ const● (C₁ [ Thunk n₁ σ₁ ])
    lem = ≡-sym (⁻¹-const●-lemma C C₁)
    r   = gpayResult∋       C'' m {n = n₂} {σ = σ₁} Σ₁ x₁ lem
    rs  = gpayResult∋-sound C'' m {n = n₂} {σ = σ₁} Σ₁ x₁ lem

  gpayResult∋-gwait-sound C m n n₁ .(n + n₁) Σ₁ x₁ (≜● {σ = σ}) =
    record
      { Δ-eq  = GPaySound.Δ-eq  rs
      ; ρ-eq  = GPaySound.ρ-eq  rs
      ; Σ₂-eq = GPaySound.Σ₂-eq rs
      ; v₂-eq = cast-lemma₁₀
                  (C ⁻¹∘ ●) (C ⁻¹∘ ●)
                  n (n ∸ (m ∸ n₁)) x₁ (GPay.v₂ r)
                  (GPaySound.Δ-eq rs)
                  (GPaySound.Δ-eq rs)
                  (erase-lemma (C ⁻¹∘ ●))
                  (erase-lemma (C ⁻¹∘ ●))
                  ≡-refl
                  (≜-⁻¹-lemma C ≜●)
                  (≡-cong (\i -> C ⁻¹ [ Thunk i σ ])
                          (i∸k∸j+j∸k≡i+j∸k n n₁ m))
                  ≡-refl
                  (GPaySound.v₂-eq rs)
      }
    where
    C'  = C ⁻¹
    r   = gpayResult∋       C' m Σ₁ x₁ ≡-refl
    rs  = gpayResult∋-sound C' m Σ₁ x₁ ≡-refl

  gpayResult∋-gwait-sound C m n n₁ n₂ Σ₁ x₁ (≜●T {C = C'}) =
    record
      { Δ-eq  = GPaySound.Δ-eq  rs
      ; ρ-eq  = GPaySound.ρ-eq  rs
      ; Σ₂-eq = GPaySound.Σ₂-eq rs
      ; v₂-eq = cast-lemma₁₀
                  (C ⁻¹∘ ●) (C ⁻¹∘ ●)
                  n n x₁ (GPay.v₂ r)
                  (GPaySound.Δ-eq rs)
                  (GPaySound.Δ-eq rs)
                  (erase-lemma C'')
                  (erase-lemma (C ⁻¹∘ Thunk● (n + n₁) C'))
                  (⁻¹_[]-lemma C (Thunk● n₁ C'))
                  (≜-⁻¹-lemma C ≜●T)
                  (⁻¹_[]-lemma C (Thunk● (n + n₁) C'))
                  (≡-sym (⁻¹_[]-lemma C (Thunk● n₁ C')))
                  (GPaySound.v₂-eq rs)
      }
    where
    C'' = C ⁻¹∘ Thunk● n₁ C'
    lem = ⁻¹_[]-lemma C (Thunk● n₁ C')
    r   = gpayResult∋       C'' m Σ₁ x₁ lem
    rs  = gpayResult∋-sound C'' m Σ₁ x₁ lem

  gpayResult∋-gwait-sound C m n n₁ n₂ Σ₁ x₁ (≜T● {C = C'}) =
    record
      { Δ-eq  = GPaySound.Δ-eq  rs
      ; ρ-eq  = GPaySound.ρ-eq  rs
      ; Σ₂-eq = GPaySound.Σ₂-eq rs
      ; v₂-eq = cast-lemma₁₀
                  (C ⁻¹∘ Thunk● n₂ C') (C ⁻¹∘ Thunk● (n₂ ∸ m) C')
                  n n x₁ (GPay.v₂ r)
                  (GPaySound.Δ-eq rs)
                  (GPaySound.Δ-eq rs)
                  (erase-lemma (C ⁻¹∘ ●))
                  (erase-lemma (C ⁻¹∘ ●))
                  (≡-sym (⁻¹_[]-lemma C (Thunk● n₂ C')))
                  (≜-⁻¹-lemma C ≜T●)
                  (≡-sym (⁻¹_[]-lemma C (Thunk● (n₂ ∸ m) C')))
                  (⁻¹_[]-lemma C (Thunk● (n₂ ∸ m) C'))
                  (GPaySound.v₂-eq rs)
      }
    where
    C'' = C ⁻¹
    lem = ≡-sym (⁻¹_[]-lemma C (Thunk● n₂ C'))
    r   = gpayResult∋       C'' m Σ₁ x₁ lem
    rs  = gpayResult∋-sound C'' m Σ₁ x₁ lem

  gpayResult∋-gwait-sound C m n n₁ n₂ Σ₁ x₁ (≜TT {n = n'} eq) =
    irr (≜-⁻¹-lemma (Thunk● n' C) eq) (≜-⁻¹-lemma C (≜TT eq))
        (gpayResult∋-gwait-sound (Thunk● n' C) m n n₁ n₂ Σ₁ x₁ eq)

  gpayResult∋-gwait-sound C m n n₁ n₂ Σ₁ x₁ (≜●×² {σ = σ} eq) =
    irr (≜-⁻¹-lemma (C ●× σ) eq) (≜-⁻¹-lemma C (≜●×² eq))
      (gpayResult∋-gwait-sound (C ●× σ) m n n₁ n₂ Σ₁ x₁ eq)

  gpayResult∋-gwait-sound C m n n₁ n₂ Σ₁ x₁ (≜×●² {σ = σ} eq) =
    irr (≜-⁻¹-lemma (σ ×● C) eq) (≜-⁻¹-lemma C (≜×●² eq))
      (gpayResult∋-gwait-sound (σ ×● C) m n n₁ n₂ Σ₁ x₁ eq)

  gpayResult∋-gwait-sound C m n n₁ n₂ Σ₁ x₁
    (≜●××● {C₁ = C₁} {C₂ = C₂} {σ₁ = σ₁} {σ₂ = σ₂}) = record
      { Δ-eq  = GPaySound.Δ-eq  rs
      ; ρ-eq  = GPaySound.ρ-eq  rs
      ; Σ₂-eq = GPaySound.Σ₂-eq rs
      ; v₂-eq = cast-lemma₁₀
                  (C ⁻¹∘ (C₁ ●× C₂ [ Thunk n₂ σ₂ ]))
                  (C ⁻¹∘ (C₁ ●× C₂ [ Thunk (n₂ ∸ m) σ₂ ]))
                  n n x₁ (GPay.v₂ r)
                  (GPaySound.Δ-eq rs)
                  (GPaySound.Δ-eq rs)
                  (erase-lemma C')
                  (erase-lemma (C ⁻¹∘ (C₁ [ Thunk (n + n₁) σ₁ ] ×● C₂)))
                  (≡-sym (×-lemma C C₁ C₂))
                  (≜-⁻¹-lemma C ≜●××●)
                  (≡-sym (×-lemma C C₁ C₂))
                  (×-lemma C C₁ C₂)
                  (GPaySound.v₂-eq rs)
      }
    where
    C'  = C ⁻¹∘ (C₁ [ Thunk n₁ σ₁ ] ×● C₂)
    lem = ≡-sym (×-lemma C C₁ C₂)
    r   = gpayResult∋       C' m Σ₁ x₁ lem
    rs  = gpayResult∋-sound C' m Σ₁ x₁ lem

  gpayResult∋-gwait-sound C m n n₁ n₂ Σ₁ x₁
    (≜×●●× {C₁ = C₁} {C₂ = C₂} {σ₁ = σ₁} {σ₂ = σ₂}) = record
      { Δ-eq  = GPaySound.Δ-eq  rs
      ; ρ-eq  = GPaySound.ρ-eq  rs
      ; Σ₂-eq = GPaySound.Σ₂-eq rs
      ; v₂-eq = cast-lemma₁₀
                  (C ⁻¹∘ (C₁ [ Thunk n₂ σ₁ ] ×● C₂))
                  (C ⁻¹∘ (C₁ [ Thunk (n₂ ∸ m) σ₁ ] ×● C₂))
                  n n x₁ (GPay.v₂ r)
                  (GPaySound.Δ-eq rs)
                  (GPaySound.Δ-eq rs)
                  (erase-lemma C')
                  (erase-lemma (C ⁻¹∘ (C₁ ●× C₂ [ Thunk (n + n₁) σ₂ ])))
                  (×-lemma C C₁ C₂)
                  (≜-⁻¹-lemma C ≜×●●×)
                  (×-lemma C C₁ C₂)
                  (≡-sym (×-lemma C C₁ C₂))
                  (GPaySound.v₂-eq rs)
      }
    where
    C'  = C ⁻¹∘ (C₁ ●× C₂ [ Thunk n₁ σ₂ ])
    lem = ×-lemma C C₁ C₂
    r   = gpayResult∋       C' m Σ₁ x₁ lem
    rs  = gpayResult∋-sound C' m Σ₁ x₁ lem

gpayResult-sound :  forall {Γ} C m (Σ₁ : Heap Γ) {n σ}
                           (v₁ : Γ ⊢⇓ C [ Thunk n σ ])
                 -> GPaySound _⊢⇓_ ⌜_⌝⇓ (gpayResult C m Σ₁ v₁) Σ₁ v₁
gpayResult-sound {Γ = Γ} ● m Σ₁ (return⋆⇓ v₁) = record
  { Δ-eq  = ≡-refl
  ; ρ-eq  = ⌜id⌝-lemma Γ
  ; Σ₂-eq = ≡-refl
  ; v₂-eq = ≡-refl
  }

gpayResult-sound {Γ = Γ} (const● σ) m Σ₁ v₁ = record
  { Δ-eq  = ≡-refl
  ; ρ-eq  = ⌜id⌝-lemma Γ
  ; Σ₂-eq = ≡-refl
  ; v₂-eq = ≡-refl
  }

gpayResult-sound {Γ = Γ} (C ●× σ₂) m Σ₁ {n = n} {σ = σ} (x₁ ,⇓ y₁) =
  record
    { Δ-eq  = Δ-eq
    ; ρ-eq  = GPaySound.ρ-eq  rs
    ; Σ₂-eq = GPaySound.Σ₂-eq rs
    ; v₂-eq = begin
        cast₁ (cast₂ (⌜ x₁ ⌝∋ ,⇓ ⌜ y₁ ⌝∋))
                                                ≡⟨ cast-lemma₁ ⌜ x₁ ⌝∋ ⌜ y₁ ⌝∋ Δ-eq lem ⟩
        cast₄ (cast₅ ⌜ x₁ ⌝∋) ,⇓ cast₃ ⌜ y₁ ⌝∋
                                                ≡⟨ ≡-cong₂ _,⇓_ (GPaySound.v₂-eq rs)
                                                                (/∋-ρ-lemma rs y₁) ⟩
        ⌜ GPay.v₂ r ⌝∋ ,⇓ ⌜ y₁ /∋ GPay.ρ r ⌝∋
                                                ∎
    }
  where
  r     = gpayResult∋ C m Σ₁ x₁ ≡-refl
  rs    = gpayResult∋-sound C m Σ₁ x₁ ≡-refl
  Δ-eq  = GPaySound.Δ-eq rs
  lem   = erase-lemma C
  lem'  = erase-lemma (C ●× σ₂)
  χ     = ⌜ C [ Thunk (n ∸ m) σ ] ⌝⋆
  cast₁ = ≡-subst (\Δ -> Δ ⊢⇓ χ × ⌜ σ₂ ⌝⋆) Δ-eq
  cast₂ = ≡-subst (_⊢⇓_ ⌜ Γ ⌝▻) lem'
  cast₃ = ≡-subst (\Γ -> Γ ∋ ⌜ σ₂ ⌝⋆) Δ-eq
  cast₄ = ≡-subst (\Γ -> Γ ∋ χ) Δ-eq
  cast₅ = ≡-subst (_∋_ ⌜ Γ ⌝▻) lem

gpayResult-sound {Γ = Γ} (σ₁ ×● C) m Σ₁ {n = n} {σ = σ} (x₁ ,⇓ y₁) =
  record
    { Δ-eq  = GPaySound.Δ-eq  rs
    ; ρ-eq  = GPaySound.ρ-eq  rs
    ; Σ₂-eq = GPaySound.Σ₂-eq rs
    ; v₂-eq = begin
        cast₁ (cast₂ (⌜ x₁ ⌝∋ ,⇓ ⌜ y₁ ⌝∋))
                                                ≡⟨ cast-lemma₂ ⌜ x₁ ⌝∋ ⌜ y₁ ⌝∋ Δ-eq lem ⟩
        cast₃ ⌜ x₁ ⌝∋ ,⇓ cast₄ (cast₅ ⌜ y₁ ⌝∋)
                                                ≡⟨ ≡-cong₂ _,⇓_ (/∋-ρ-lemma rs x₁)
                                                                (GPaySound.v₂-eq rs) ⟩
        ⌜ x₁ /∋ GPay.ρ r ⌝∋ ,⇓ ⌜ GPay.v₂ r ⌝∋
                                                ∎
    }
  where
  r     = gpayResult∋ C m Σ₁ y₁ ≡-refl
  rs    = gpayResult∋-sound C m Σ₁ y₁ ≡-refl
  Δ-eq  = GPaySound.Δ-eq rs
  lem   = erase-lemma C
  lem'  = erase-lemma (σ₁ ×● C)
  χ     = ⌜ C [ Thunk (n ∸ m) σ ] ⌝⋆
  cast₁ = ≡-subst (\Δ -> Δ ⊢⇓ ⌜ σ₁ ⌝⋆ × χ) Δ-eq
  cast₂ = ≡-subst (_⊢⇓_ ⌜ Γ ⌝▻) lem'
  cast₃ = ≡-subst (\Γ -> Γ ∋ ⌜ σ₁ ⌝⋆) Δ-eq
  cast₄ = ≡-subst (\Γ -> Γ ∋ χ) Δ-eq
  cast₅ = ≡-subst (_∋_ ⌜ Γ ⌝▻) lem

gpayResult-sound (Thunk● k C) m Σ₁ (return⋆⇓ v₁) = record
  { Δ-eq  = GPaySound.Δ-eq  rs
  ; ρ-eq  = GPaySound.ρ-eq  rs
  ; Σ₂-eq = GPaySound.Σ₂-eq rs
  ; v₂-eq = GPaySound.v₂-eq rs
  } where rs = gpayResult-sound C m Σ₁ v₁
