------------------------------------------------------------------------
-- PayCtxt-related functions
------------------------------------------------------------------------

module PayCtxt where

open import Kind
open import TypeSystem

open import Relation.Binary.PropositionalEquality
  using (_≡_; ≡-refl; ≡-cong)

------------------------------------------------------------------------
-- Composing contexts

infixr 42 _∘_

_∘_ : PayCtxt -> PayCtxt -> PayCtxt
●           ∘ C₂ = C₂
const● σ    ∘ C₂ = const● σ
(C₁ ●× σ)   ∘ C₂ = (C₁ ∘ C₂) ●× σ
(σ ×● C₁)   ∘ C₂ = σ ×● (C₁ ∘ C₂)
Thunk● n C₁ ∘ C₂ = Thunk● n (C₁ ∘ C₂)

∘-lemma : forall C₁ C₂ {σ} -> C₁ ∘ C₂ [ σ ] ≡ C₁ [ C₂ [ σ ] ]
∘-lemma ●             C₂ = ≡-refl
∘-lemma (const● σ)    C₂ = ≡-refl
∘-lemma (C₁ ●× σ)     C₂ = ≡-cong (\τ -> τ × σ) (∘-lemma C₁ C₂)
∘-lemma (σ ×● C₁)     C₂ = ≡-cong (\τ -> σ × τ) (∘-lemma C₁ C₂)
∘-lemma (Thunk● n C₁) C₂ = ≡-cong (Thunk n) (∘-lemma C₁ C₂)

------------------------------------------------------------------------
-- "Inverting" contexts

-- "Inverting" here means swapping the order of the context elements,
-- so that the outermost context constructor is the first one applied
-- to the hole-filling type. In other words, the context is turned
-- outside-in.

-- Instantiates a context outside-in.

infix 41 _⁻¹[_]

_⁻¹[_] : PayCtxt -> Ty thunked -> Ty thunked
●          ⁻¹[ τ ] = τ
const● σ   ⁻¹[ τ ] = σ
(C ●× σ)   ⁻¹[ τ ] = C ⁻¹[ τ × σ ]
(σ ×● C)   ⁻¹[ τ ] = C ⁻¹[ σ × τ ]
Thunk● n C ⁻¹[ τ ] = C ⁻¹[ Thunk n τ ]

-- "Inverts" a context and composes the result with another.

infixr 42 _⁻¹∘_

_⁻¹∘_ : PayCtxt -> PayCtxt -> PayCtxt
●           ⁻¹∘ C₂ = C₂
const● σ    ⁻¹∘ C₂ = const● σ
(C₁ ●× σ)   ⁻¹∘ C₂ = C₁ ⁻¹∘ (C₂ ●× σ)
(σ ×● C₁)   ⁻¹∘ C₂ = C₁ ⁻¹∘ (σ ×● C₂)
Thunk● n C₁ ⁻¹∘ C₂ = C₁ ⁻¹∘ Thunk● n C₂

-- "Inverts" a context.

infix 43 _⁻¹

_⁻¹ : PayCtxt -> PayCtxt
C ⁻¹ = C ⁻¹∘ ●

-- The definitions are closely related.

⁻¹[]-lemma : forall C C' {σ} -> C ⁻¹[ C' [ σ ] ] ≡ C ⁻¹∘ C' [ σ ]
⁻¹[]-lemma ●            C' = ≡-refl
⁻¹[]-lemma (const● σ)   C' = ≡-refl
⁻¹[]-lemma (C ●× σ)     C' = ⁻¹[]-lemma C (C' ●× σ)
⁻¹[]-lemma (σ ×● C)     C' = ⁻¹[]-lemma C (σ ×● C')
⁻¹[]-lemma (Thunk● n C) C' = ⁻¹[]-lemma C (Thunk● n C')

⁻¹-lemma : forall C {σ} -> C ⁻¹[ σ ] ≡ C ⁻¹ [ σ ]
⁻¹-lemma C = ⁻¹[]-lemma C ●

⁻¹∘-∘-assoc : forall C₁ C₂ C₃ -> C₁ ⁻¹∘ (C₂ ∘ C₃) ≡ (C₁ ⁻¹∘ C₂) ∘ C₃
⁻¹∘-∘-assoc ●             C₂ C₃ = ≡-refl
⁻¹∘-∘-assoc (const● σ)    C₂ C₃ = ≡-refl
⁻¹∘-∘-assoc (C₁ ●× σ)     C₂ C₃ = ⁻¹∘-∘-assoc C₁ (C₂ ●× σ) C₃
⁻¹∘-∘-assoc (σ ×● C₁)     C₂ C₃ = ⁻¹∘-∘-assoc C₁ (σ ×● C₂) C₃
⁻¹∘-∘-assoc (Thunk● n C₁) C₂ C₃ = ⁻¹∘-∘-assoc C₁ (Thunk● n C₂) C₃

⁻¹∘-lemma : forall C₁ C₂ -> C₁ ⁻¹∘ C₂ ≡ C₁ ⁻¹ ∘ C₂
⁻¹∘-lemma C₁ C₂ = ⁻¹∘-∘-assoc C₁ ● C₂
