------------------------------------------------------------------------
-- Some substitution-related lemmas
------------------------------------------------------------------------

module Substitution.Lemmas where

open import Kind
open import TypeSystem
open import Substitution
open VarSubst
open import OperationalSemantics.Binding

open import Relation.Binary.PropositionalEquality
open ≡-Reasoning

------------------------------------------------------------------------
-- The identity substitution is really an identity (for simple
-- variables)

wk⇒-vs-lemma :  forall {Γ Δ : Ctxt simple} {σ τ}
                       (x : Γ ∋ σ) (ρ : Γ ⇒ Δ)
             -> x /∋ wk⇒ {σ = τ} ρ ≡ vs (x /∋ ρ)
wk⇒-vs-lemma vz             (ρ ▷ y) = ≡-refl
wk⇒-vs-lemma (vs {σ = σ} x) (ρ ▷ y) = begin
  x /∋ wk⇒ ρ
                        ≡⟨ wk⇒-vs-lemma x ρ ⟩
  vs (x /∋ ρ)
                        ≡⟨ byDef ⟩
  vs (vs x /∋ (ρ ▷ y))
                        ∎

/∋-id-lemma : forall {Γ : Ctxt simple} {σ} (x : Γ ∋ σ) -> x /∋ idˢ ≡ x
/∋-id-lemma vz     = ≡-refl
/∋-id-lemma (vs x) = begin
  vs x /∋ idˢ
                 ≡⟨ byDef ⟩
  x /∋ wk⇒ idˢ
                 ≡⟨ wk⇒-vs-lemma x idˢ ⟩
  vs (x /∋ idˢ)
                 ≡⟨ ≡-cong vs (/∋-id-lemma x) ⟩
  vs x
                 ∎

∘-id-lemma : forall {Γ Δ : Ctxt simple} (ρ : Γ ⇒ Δ) -> ρ ∘ˢ idˢ ≡ ρ
∘-id-lemma ∅       = ≡-refl
∘-id-lemma (ρ ▷ x) = ≡-cong₂ _▷_ (∘-id-lemma ρ) (/∋-id-lemma x)

/⊢-id-lemma : forall {Γ : Ctxt simple} {σ} (e : Γ ⊢ σ) -> e /⊢ idˢ ≡ e
/⊢-id-lemma (var x)        = ≡-cong var (/∋-id-lemma x)
/⊢-id-lemma (λ t)          = ≡-cong λ_ (/⊢-id-lemma t)
/⊢-id-lemma (λ² t)         = ≡-cong λ²_ (/⊢-id-lemma t)
/⊢-id-lemma (t₁ · t₂)      = ≡-cong₂ _·_ (/⊢-id-lemma t₁) (/⊢-id-lemma t₂)
/⊢-id-lemma (fix t)        = ≡-cong fix (/⊢-id-lemma t)
/⊢-id-lemma z              = ≡-refl
/⊢-id-lemma s              = ≡-refl
/⊢-id-lemma (natrec t₁ t₂) = ≡-cong₂ natrec (/⊢-id-lemma t₁) (/⊢-id-lemma t₂)
/⊢-id-lemma (t₁ , t₂)      = ≡-cong₂ _,_ (/⊢-id-lemma t₁) (/⊢-id-lemma t₂)
/⊢-id-lemma (uncurry t)    = ≡-cong uncurry (/⊢-id-lemma t)

/↦-id-lemma : forall {Γ : Ctxt simple} {σ} (b : Γ ↦ σ) -> b /↦ idˢ ≡ b
/↦-id-lemma (↦ⁿˢ _ e) = ≡-cong ↦_ (/⊢-id-lemma e)

------------------------------------------------------------------------
-- Aside: For annotated variables the identity is not an identity

id-not-identity = begin
  vs (gwait∋ C m vz) /∋ idˢ {Γ = Γ}
                                     ≡⟨ byDef ⟩
  gwait∋ C m vz /∋ wk
                                     ≡⟨ byDef ⟩
  gwait∋ C m (vz /∋ wk)
                                     ≡⟨ byDef ⟩
  gwait∋ C m (vs vz)
                                     ∎
  where
  C = ●
  m = 0
  Γ = ε ▻ Thunk 0 Nat ▻ Nat

-- I would say that the problem here is not really the annotated
-- variables, but the use of the propositional equality _≡_. A weaker
-- equality, which for instance allowed permuting gwait∋ and vs, would
-- probably get around this issue. Another option is to force all
-- gwait∋ to be on the "outside". I used such an approach before, but
-- switched to the current one because it was a little easier to work
-- with.

-- Note that it is sound to extend the thunked language with arbitrary
-- constructions, as long as the main proofs go through, so this issue
-- is not important.
